#include <allegro.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "global.h"
#include "palette.h"
#include "texture.h"
#include <vector>
#include <iostream>
#include <fstream>

using namespace std;


float fDepth;
float fPresistance;
float fScale;
int   nOctaves;

vector<CTexture *> textures;
vector<CTexture *>::iterator currTexIt;
CTexture * pCurrTex;

const int bkgCol = RGB(155,155,155);
const int nGap = 10;

void ShowInfo(int x, int y, CTexture * tex)
{
    const int step = 9;

    textprintf(scrbuf, font, x, y - step,      RGB(255,255,0  ), "[TAB]        Next Texture");
    textprintf(scrbuf, font, x, y + step,      RGB(255,255,255), "[HOME/END]   Freq. start %.2f ", tex->m_fFreqStart);
    textprintf(scrbuf, font, x, y + step * 2,  RGB(255,255,255), "[LEFT/RIGHT] Freq. mod   %.2f ", tex->m_fFreqMod);
    textprintf(scrbuf, font, x, y + step * 3,  RGB(255,255,255), "[</>]        Ampl. start %.2f ", tex->m_fAmplStart);
    textprintf(scrbuf, font, x, y + step * 4,  RGB(255,255,255), "['/?]        Ampl. mod   %.2f ", tex->m_fAmplMod);
    textprintf(scrbuf, font, x, y + step * 5,  RGB(255,255,255), "[UP/DOWN]    Depth       %.2f ", tex->m_fDepth);
    textprintf(scrbuf, font, x, y + step * 6,  RGB(255,255,255), "[PGUP/PGDN]  Octaves     %.2d ", tex->m_nOctaves);
    textprintf(scrbuf, font, x, y + step * 7,  RGB(255,255,255), "[S/X]        ScaleX      %.2f ", tex->m_fScaleX);
    textprintf(scrbuf, font, x, y + step * 8,  RGB(255,255,255), "[D/C]        ScaleY      %.2f ", tex->m_fScaleY);
    textprintf(scrbuf, font, x, y + step * 9,  RGB(255,255,255), "[A/Z]        Scale");
    if (tex->m_bTurbulence)
        textprintf(scrbuf, font, x, y + step * 10, RGB(255,255,255), "[N/T]        Mode        Turbulence");
    else
        textprintf(scrbuf, font, x, y + step * 10, RGB(255,255,255), "[N/T]        Mode        Noise");

    textprintf(scrbuf, font, x, y + step * 12, RGB(255,255,255), "F1 - Raw noise, F2 - Wood, F3 - Marble, F6 - Pattern");
    textprintf(scrbuf, font, x, y + step * 14, RGB(255,255,255), "Hold right CTRL to show and edit palette");
}

volatile g_nTime = 0;
volatile g_nMsgTime = 5;

void TimerProc(void)
{
    g_nTime++;
    g_nMsgTime++;
}
END_OF_FUNCTION(TimerProc);

char msg[256];
bool bShowMsg = false;

void ShowMsg()
{
    if (g_nMsgTime < 25) 
        textprintf(scrbuf, font, 5, nScreenHeight - font->height - 5, RGB(255,255,255), msg);
}

int main(void)
{
    allegro_init();
    set_window_title("Noise generator v.1.0");
    install_keyboard(); 
    install_timer();

    LOCK_VARIABLE(g_nTime);
    LOCK_FUNCTION(TimerProc);

    install_int_ex(TimerProc, MSEC_TO_TIMER(50));

    set_color_depth(32);
    if (set_gfx_mode(GFX_DIRECTX_WIN, nScreenWidth, nScreenHeight, 0, 0) != 0) 
        return 1;

    textprintf_centre(screen, font, nScreenWidth / 2, nScreenHeight / 2, RGB(255,255,255), "Initializing...");

    GenRandomNumbers();

    CTexture texture0(256,256,256,256);
    CTexture texture1(256,256,256,256);
    CTexture outTex(256, 256, 256, 256);

    textures.push_back(&texture0);
    textures.push_back(&texture1);

    texture0.m_pPalette->SetRGB(150, 0, 0, 0, true);
    texture0.m_pPalette->SetRGB(0, 173, 124, 22, true);
    texture0.m_pPalette->SetRGB(255, 186, 157 , 44, true);
    texture0.m_pPalette->Make();

    texture1.m_pPalette->SetRGB(0, 0, 0, 0, true);
    texture1.m_pPalette->SetRGB(67, 40, 0, 0, true);
    texture1.m_pPalette->SetRGB(92, 244, 0, 0, true);
    texture1.m_pPalette->SetRGB(115, 0, 0 , 0, true);
    texture1.m_pPalette->SetRGB(161, 0, 0 , 0, true);
    texture1.m_pPalette->SetRGB(170, 228, 232 , 0, true);
    texture1.m_pPalette->SetRGB(175, 0, 0 , 0, true);
    texture1.m_pPalette->SetRGB(255, 0, 0 , 0, true);
    texture1.m_pPalette->Make();


    scrbuf = create_bitmap(nScreenWidth, nScreenHeight);

    text_mode(-1);


    fDepth = 0.5;
    fPresistance = 0.5;
    fScale = 50.0;
    nOctaves = 5;

    texture0.SetAmplitude(1.0, 0.65);
    texture0.SetFrequency(1.0, 2.0);
    texture0.SetScale(250.0, 250.0, 250.0);
    texture0.SetParams(0.65, 1.0, 5);

    texture1.SetAmplitude(1.0, 0.65);
    texture1.SetFrequency(1.0, 2.0);
    texture1.SetScale(200.0, 200.0, 200.0);
    texture1.SetParams(0.65, 1.0, 5);

    outTex.SetAmplitude(1.0, 0.65);
    outTex.SetFrequency(1.0, 2.0);
    outTex.SetScale(200.0, 200.0, 200.0);
    outTex.SetParams(0.65, 1.0, 5);

    texture1.Generate();

    pCurrTex = &texture0;
    currTexIt = textures.begin();

    bool bShowpalette = false;

    while (!key[KEY_ESC])
    {
        clear_to_color(scrbuf, bkgCol);
        acquire_screen();

        if (!key[KEY_RCONTROL] && !key[KEY_LCONTROL])
        {        
            if (bShowpalette)
            {
                pCurrTex->m_bChange = true;
                bShowpalette = false;
            }
            if (key[KEY_TAB])
            {
                WAIT(KEY_TAB);
                currTexIt++;
                if (currTexIt == textures.end())
                    currTexIt = textures.begin();
                pCurrTex = (CTexture *) *currTexIt;
            }
            if (key[KEY_UP])    pCurrTex->AdjustDepth(0.02);
            if (key[KEY_DOWN])  pCurrTex->AdjustDepth(-0.02);
            if (key[KEY_QUOTE]) pCurrTex->AdjustAmplStart(0.05);
            if (key[KEY_SLASH]) pCurrTex->AdjustAmplStart(-0.05);
            if (key[KEY_COMMA]) pCurrTex->AdjustAmplMod(0.05);
            if (key[KEY_STOP])  pCurrTex->AdjustAmplMod(-0.05);
            if (key[KEY_HOME])  pCurrTex->AdjustFreqStart(0.05);
            if (key[KEY_END])   pCurrTex->AdjustFreqStart(-0.05);
            if (key[KEY_RIGHT]) pCurrTex->AdjustFreqMod(0.05);
            if (key[KEY_LEFT])  pCurrTex->AdjustFreqMod(-0.05);
            if (key[KEY_PGUP])  pCurrTex->AdjustOctaves(1);
            if (key[KEY_PGDN])  pCurrTex->AdjustOctaves(-1);
            if (key[KEY_A])
            {
                pCurrTex->AdjustScaleX(5.0);
                pCurrTex->AdjustScaleY(5.0);
                pCurrTex->AdjustScaleZ(5.0);
            }
            if (key[KEY_Z])
            {
                pCurrTex->AdjustScaleX(-5.0);
                pCurrTex->AdjustScaleY(-5.0);
                pCurrTex->AdjustScaleZ(-5.0);
            }
            if (key[KEY_S]) pCurrTex->AdjustScaleX(5.0);
            if (key[KEY_X]) pCurrTex->AdjustScaleX(-5.0);
            if (key[KEY_D]) pCurrTex->AdjustScaleY(5.0);
            if (key[KEY_C]) pCurrTex->AdjustScaleY(-5.0);

            if (key[KEY_F1]) pCurrTex->SetMode(MODE_NOISE);
            if (key[KEY_F2]) pCurrTex->SetMode(MODE_WOOD);
            if (key[KEY_F3]) pCurrTex->SetMode(MODE_MARBLE);
            if (key[KEY_F4]) pCurrTex->SetMode(MODE_PATTERN);
            if (key[KEY_F5])
            {
                //pCurrTex->SetMode(MODE_BMP);
                //pCurrTex->m_pTexture = load_tga("texture.tga", NULL);
            }
            if (key[KEY_F6]) pCurrTex->SetMode(MODE_WOOD2);
            if (key[KEY_N])
            {
                pCurrTex->m_bTurbulence = false;
                pCurrTex->m_bChange = true;
            }
            if (key[KEY_T])
            {
                pCurrTex->m_bTurbulence = true;
                pCurrTex->m_bChange = true;
            }

        }
        
        if (key[KEY_RCONTROL])
        {
            pCurrTex->m_pPalette->Show(scrbuf, nScreenWidth / 2 - nCols / 2, nScreenHeight - 20 - 5 ,20);
            pCurrTex->m_pPalette->HandleMenu(scrbuf);
            bShowpalette = true;
//            pCurrTex->m_bChange = true;
        }

        if (key[KEY_LCONTROL])
        {
            if (key[KEY_S])
            {
                g_nMsgTime = 0;
                strcpy(msg, "File saved (texture0.info)...");
                WAIT(KEY_S);
                pCurrTex->Save("texture0.info");
            }

            if (key[KEY_L])
            {
                g_nMsgTime = 0;
                

            }
        }

        ShowInfo(nScreenWidth / 2 - 150, 0 + 256 + 10 * 3, pCurrTex);

        clear_keybuf();
        vector<CTexture *>::iterator it;
        int nTextures = textures.size() + 1;
        int nPosX = nScreenWidth / 2 - ((nTextures - 1) * nGap + nTextures * nWidth) / 2;
        int nPosY = 10;
        for (it = textures.begin(); it != textures.end(); it++)
        {
            CTexture * tex = (CTexture *) *it;
            int col = 0;
            if (pCurrTex == tex)
            {
                tex->Generate();
                col = RGB(255,255,0);
            }
            tex->Show(scrbuf, nPosX, nPosY, col);
            nPosX += nWidth + nGap;
        }
        outTex.Set(&texture0);
        //outTex.Mul(&texture1);
        //outTex.Mix(&texture1);
        outTex.Show(scrbuf, nPosX, nPosY, RGB(255, 255, 255));

        ShowMsg();

        blit(scrbuf, screen, 0, 0, 0, 0, nScreenWidth, nScreenHeight);

        release_screen();

    }

    destroy_bitmap(scrbuf);
    return 0;
}

END_OF_MAIN();
