#include <allegro.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include "global.h"
#include "palette.h"


float randTab[nWidth * nHeight * nDepth];

BITMAP * scrbuf;
BITMAP * texbuf;

#define FLOOR(x) ((int)(x) - ((x) < 0 && (x) != (int)(x)))
#define A -0.75       // constraint parameter

/*-------------------------------------------------------------------------------------*/
float Clamp(float x, float min, float max)
{
    return (x < min ? min : (x > max ? max : x));
}
void ClampRGB(CPalette::colRGB & col, float min, float max)
{
    if (col.r < min) col.r = min; else if (col.r > max) col.r = max;
    if (col.g < min) col.g = min; else if (col.g > max) col.g = max;
    if (col.b < min) col.b = min; else if (col.b > max) col.b = max;
}
/*-------------------------------------------------------------------------------------*/
inline float Noise(int x, int y)
{
    x = x & (nWidth - 1);
    y = y & (nHeight - 1);
    return randTab[y * nWidth + x] / 12.0;
}
inline float Noise(int x, int y, int z)
{
    x = x & (nWidth - 1);
    y = y & (nHeight - 1);
    z = z & (nDepth - 1);
    return randTab[(z * nWidth * nHeight) + y * nWidth + x] / 1.0;
}
/*-------------------------------------------------------------------------------------*/
inline float Cubic(float x, float a)
{
    float w;

    if ((x >= 0.0) && (x < 1.0)) 
        w = ((a + 2.0) * x - (a + 3.0)) * x * x + 1.0;
    else if ((x >= 1.0) && (x <= 2.0)) 
        w = ((a * x - 5.0 * a) * x + 8.0 * a) * x - 4.0 * a;
    else 
        w = 0.0;
    return w;
}

inline float InterpolateCubic2(float v0, float v1, float v2, float v3, float x)
{
    return v0 * Cubic(1.0 + x, A) +
           v1 * Cubic(0.0 + x, A) +
           v2 * Cubic(1.0 - x, A) +
           v3 * Cubic(2.0 - x, A);
}

inline float InterpolateCubic(float v0, float v1,float v2, float v3, float x)
{
    float v4 = v0 - v1;
    float P  = v3 - v2 - v4;
    float Q  = v4 - P;
    float R  = v2 - v0;
    return (x * (P * x * x + Q * x + R) + v1);
}
float InterpolateCubic(float *i, float x)
{
    float x1, x2, x3;
    float h1, h2, h3, h4;
    float result;

    x1 = x + 1.0;
    x2 = x1 * x1;
    x3 = x2 * x1;
    h4 = -x3 + 5 * x2 - 8 * x1 + 4;
    result = h4 * i[0];

    x1 = x;
    x2 = x1 * x1;
    x3 = x2 * x1;
    h1 = x3 - 2 * x2 + 1;
    result += h1 * i[1];

    x1 = 1.0 - x;
    x2 = x1 * x1;
    x3 = x2 * x1;
    h2 = x3 - 2 * x2 + 1;
    result += h2 * i[2];

    x1 = 2.0 - x;
    x2 = x1 * x1;
    x3 = x2 * x1;
    h3 = -x3 + 5 * x2 - 8 * x1 + 4;
    result += h3 * i[3];

    return result;
}

inline float InterpolateCubic(int n, float x)
{
    float v4 = randTab[n-1] - randTab[n-0];
    float P  = randTab[n+2] - randTab[n+1] - v4;
    float Q  = v4 - P;
    float R  = randTab[n+1] - randTab[n-1];
    return x * (P * x * x + Q * x + R) + randTab[n-0];
}

inline float InterpolateLinear(float a, float b, float x)
{
    return  a * (1.0-x) + b * x;
}

inline float InterpolateCosine(float a, float b, float x)
{
    float ft = x * 3.1415927;
    float f = (1.0 - cos(ft)) * 0.5;
    return  a * (1.0 - f) + b * f;
}


inline float Distance(float x0, float y0, float z0, float x1, float y1, float z1)
{
    float x = x0 - x1;
    float y = y0 - x1;
    float z = z0 - x1;
    return sqrt(x*x + y*y + z*z);
}
/*-------------------------------------------------------------------------------------*/
float InterpolatedNoiseCubic(float x, float y)
{
    int   ix = FLOOR(x);
    int   iy = FLOOR(y);
    float fx = x - ix;
    float fy = y - iy;

    float i1 = InterpolateCubic(Noise(ix-1, iy-1), Noise(ix, iy-1), Noise(ix+1, iy-1), Noise(ix+2, iy-1), fx);
    float i2 = InterpolateCubic(Noise(ix-1, iy+0), Noise(ix, iy+0), Noise(ix+1, iy+0), Noise(ix+2, iy+0), fx);
    float i3 = InterpolateCubic(Noise(ix-1, iy+1), Noise(ix, iy+1), Noise(ix+1, iy+1), Noise(ix+2, iy+1), fx);
    float i4 = InterpolateCubic(Noise(ix-1, iy+2), Noise(ix, iy+2), Noise(ix+1, iy+2), Noise(ix+2, iy+2), fx);
    float i =  InterpolateCubic(i1 , i2 , i3, i4, fy);

    return (i < -1.0 ? -1.0 : (i > 1.0 ? 1.0 : i));

}
/*-------------------------------------------------------------------------------------*/
float InterpolatedNoiseCubic(float x, float y, float z)
{
    int   ix = FLOOR(x);
    int   iy = FLOOR(y);
    int   iz = FLOOR(z);

    float fx = x - ix;
    float fy = y - iy;
    float fz = z - iz;

    float xknots[4], yknots[4], zknots[4];

    for (int k = -1; k <= 2; k++) 
    {
        for (int j = -1; j <= 2; j++) 
        {
            for (int i = -1; i <= 2; i++) 
            {
                xknots[i+1] = Noise(ix+i, iy+j, iz+k);
            }
//            yknots[j+1] = InterpolateCubic(xknots[0], xknots[1], xknots[2], xknots[3], fx);
            yknots[j+1] = InterpolateCubic(xknots, fx);
        }
        //zknots[k+1] = InterpolateCubic(yknots[0], yknots[1], yknots[2], yknots[3], fy);
        zknots[k+1] = InterpolateCubic(yknots, fy);
    }
    //float i = InterpolateCubic(zknots[0], zknots[1], zknots[2], zknots[3], fz);
    float i = InterpolateCubic(zknots, fz);

    return (i < -1.0 ? -1.0 : (i > 1.0 ? 1.0 : i));

}
/*-------------------------------------------------------------------------------------*/
float InterpolatedNoiseLinear(float x, float y, float z)
{
    int   ix = FLOOR(x);
    int   iy = FLOOR(y);
    int   iz = FLOOR(z);
    float fx = x - ix;
    float fy = y - iy;
    float fz = z - iz;

    float a, b;

    float yknots[4], zknots[4];

    for (int k = 0; k <= 1; k++) 
    {
        for (int j = 0; j <= 1; j++) 
        {
            a = Noise(ix + 0, iy + j, iz + k);
            b = Noise(ix + 1, iy + j, iz + k);

            yknots[j] = InterpolateLinear(a, b, fx);
        }
        zknots[k] = InterpolateLinear(yknots[0], yknots[1], fy);
    }
    float i = InterpolateLinear(zknots[0], zknots[1], fz);

    return (i < -1.0 ? -1.0 : (i > 1.0 ? 1.0 : i));

}
/*-------------------------------------------------------------------------------------*/
float InterpolatedNoiseCubicFast(float x, float y, float z)
{
    int   ix = FLOOR(x);
    int   iy = FLOOR(y);
    int   iz = FLOOR(z);
    float fx = x - ix;
    float fy = y - iy;
    float fz = z - iz;

    float y0, y1, y2, y3;
    float z0, z1, z2, z3;

    const int nWH = nWidth * nHeight - 1;
    int nz, ny, n;

    nz = ((iz-1) & nD) * nWidth * nHeight + ix;
    ny = ((iy-1) & nH) * nWidth; n = nz + ny;
    y0 = InterpolateCubic(n, fx);
    ny = ((iy-0) & nH) * nWidth; n = nz + ny;
    y1 = InterpolateCubic(n, fx);
    ny = ((iy+1) & nH) * nWidth; n = nz + ny;
    y2 = InterpolateCubic(n, fx);
    ny = ((iy+2) & nH) * nWidth; n = nz + ny;
    y3 = InterpolateCubic(n, fx);
    z0 = InterpolateCubic(y0, y1, y2, y3, fy);

    nz = ((iz-0) & nD) * nWidth * nHeight + ix;
    ny = ((iy-1) & nH) * nWidth; n = nz + ny;
    y0 = InterpolateCubic(n, fx);
    ny = ((iy-0) & nH) * nWidth; n = nz + ny;
    y1 = InterpolateCubic(n, fx);
    ny = ((iy+1) & nH) * nWidth; n = nz + ny;
    y2 = InterpolateCubic(n, fx);
    ny = ((iy+2) & nH) * nWidth; n = nz + ny;
    y3 = InterpolateCubic(n, fx);
    z1 = InterpolateCubic(y0, y1, y2, y3, fy);

    nz = ((iz+1) & nD) * nWidth * nHeight + ix;
    ny = ((iy-1) & nH) * nWidth; n = nz + ny;
    y0 = InterpolateCubic(n, fx);
    ny = ((iy-0) & nH) * nWidth; n = nz + ny;
    y1 = InterpolateCubic(n, fx);
    ny = ((iy+1) & nH) * nWidth; n = nz + ny;
    y2 = InterpolateCubic(n, fx);
    ny = ((iy+2) & nH) * nWidth; n = nz + ny;
    y3 = InterpolateCubic(n, fx);
    z2 = InterpolateCubic(y0, y1, y2, y3, fy);

    nz = ((iz+2) & nD) * nWidth * nHeight + ix;
    ny = ((iy-1) & nH) * nWidth; n = nz + ny;
    y0 = InterpolateCubic(n, fx);
    ny = ((iy-0) & nH) * nWidth; n = nz + ny;
    y1 = InterpolateCubic(n, fx);
    ny = ((iy+1) & nH) * nWidth; n = nz + ny;
    y2 = InterpolateCubic(n, fx);
    ny = ((iy+2) & nH) * nWidth; n = nz + ny;
    y3 = InterpolateCubic(n, fx);
    z3 = InterpolateCubic(y0, y1, y2, y3, fy);

    float i = InterpolateCubic(z0, z1, z2, z3, fz);

    return (i < -1.0 ? -1.0 : (i > 1.0 ? 1.0 : i));

}
/*-------------------------------------------------------------------------------------*/
inline float PerlinNoise(float x, float y, float persistance, int octaves)
{
    float total = 0;

    float frequency;
    float amplitude;

    for (int i = 0; i < octaves; i++)
    {
        if (i == 0)
        {
            frequency = 1.0;
            amplitude = 1.0;
        }
        else
        {
            frequency *= 2.1;
            amplitude *= persistance;
        }
        total = total + InterpolatedNoiseCubic(x * frequency, y * frequency) * amplitude;
    }
    return total;
}
/*-------------------------------------------------------------------------------------*/
float PerlinNoise(float x, float y, float z, 
                  float fFreqStart, float fFreqMod,
                  float fAmplStart, float fAmplMod,
                  int nOctaves, bool bTurbulence = false)
{
    float fTotal = 0;

    float fFrequency;
    float fAmplitude;

    for (int i = 0; i < nOctaves; i++)
    {
        if (i == 0)
        {
            fFrequency = fFreqStart;
            fAmplitude = fAmplStart;
        }
        else
        {
            fFrequency *= fFreqMod;
            fAmplitude *= fAmplMod;
        }
        float fInterNoise = InterpolatedNoiseCubicFast(x * fFrequency, y * fFrequency, z * fFrequency);
        if (bTurbulence && fInterNoise < 0.0) fInterNoise = -fInterNoise;
        fTotal = fTotal +  fInterNoise * fAmplitude;
    }
    return Clamp(fTotal, -1.0, 1.0);
}
/*-------------------------------------------------------------------------------------*/
CPalette::colRGB wood_grain(float u, float v, float w)
{
    CPalette::colRGB col;
    float radius, angle;
    int grain;

    radius = sqrt(u*u+v*v);

    if (w == 0) 
        angle = 3.1415 / 2.0;
    else
        angle = atan(u / w);

    radius = radius + 2.0 * sin(25.0 * angle + v / 3.0);
    grain = (int) (radius) % 60;

    if (grain < 15)
    {
        col.r = 45;
        col.g = 0;
        col.b = 128;
    }
    else
    { 
        col.r = 66;
        col.g = 77;
        col.b = 66;
    }
    return col;
}
/*-------------------------------------------------------------------------------------*/
void GenRandomNumbers()
{
    srand( (unsigned)time( NULL ) );

    for (int i = 0; i < nWidth * nHeight * nDepth; i++)
    {
        randTab[i] = ((float) 2.0f * (rand()/(float)RAND_MAX) - 1.0f);
    }
}
/*-------------------------------------------------------------------------------------*/
