#include <allegro.h>
#include <string.h>
#include "palette.h"
#include "global.h"

//#define textf(if (m_nColType == 0)  textprintf(bmp, font, x, y+9,  colCurrCol, "R = %.3d", m_pCols[m_nCurCol].r);


CPalette::CPalette(int nCols)
{
    m_nMinColPos = 0;
    m_nMaxColPos = 0;
    m_nColType   = 0;
    m_nCurCol    = 0;
    m_nColPos    = 0;
    m_nPrvCol    = 0;
    m_pCols = new colRGB[nCols];
    memset(m_pCols, 0, nCols * sizeof(colRGB));
    SetRGB(0, 0, 0, 0, true);
    SetRGB(nCols - 1, 255, 255, 255, true);
    m_nCols = nCols;
    m_bShowKnots = true;
}
/*-------------------------------------------------------------------------------------*/
CPalette::~CPalette()
{
    delete [] m_pCols;
}
/*-------------------------------------------------------------------------------------*/
void CPalette::Make()
{
    int rs, gs, bs;
    int re, ge, be;

    int is = 0;

    rs = m_pCols[0].r;
    gs = m_pCols[0].g;
    bs = m_pCols[0].b;

    for (int i = 1; i < m_nCols; i++)
    {
        if (m_pCols[i].bUse || i == m_nCols - 1)
        {
            re = m_pCols[i].r;
            ge = m_pCols[i].g;
            be = m_pCols[i].b;

            int nSize = i - is;

            float rd = (re - rs) / (float) nSize;
            float gd = (ge - gs) / (float) nSize;
            float bd = (be - bs) / (float) nSize;

            float r = rs;
            float g = gs;
            float b = bs;

            for (int j = is; j < i; j++)
            {
                m_pCols[j].r = (unsigned char) (r);
                m_pCols[j].g = (unsigned char) (g);
                m_pCols[j].b = (unsigned char) (b);

                r += rd;
                g += gd;
                b += bd;
            }

            is = i;

            rs = m_pCols[i].r;
            gs = m_pCols[i].g;
            bs = m_pCols[i].b;

        }
    }
}
/*-------------------------------------------------------------------------------------*/
void CPalette::Move()
{
    colRGB col = m_pCols[0];
    memcpy(&m_pCols[0], &m_pCols[1], (m_nCols - 1) * sizeof(colRGB));
    m_pCols[m_nCols - 1] = col;
}
/*-------------------------------------------------------------------------------------*/
void CPalette::Show(BITMAP * bmp, int nDrawOffX, int nDrawOffY, int nPalHeight)
{
    for (int i = 0; i < m_nCols; i++)
    {

        line(bmp, nDrawOffX + i, nDrawOffY, nDrawOffX + i, nDrawOffY + nPalHeight, RGBI(i));

        if (m_pCols[i].bUse)
        {
            int nCol = RGB(255,255,0);
            if (m_nCurCol == i)
                nCol = RGB(255,0,0);

            if (m_bShowKnots)
            {
            
                triangle(bmp, nDrawOffX + i - 3, nDrawOffY - 5,
                            nDrawOffX + i + 3, nDrawOffY - 5,
                            nDrawOffX + i    , nDrawOffY - 0,
                            nCol);
                line(bmp, nDrawOffX + i, nDrawOffY, nDrawOffX + i, nDrawOffY + nPalHeight, nCol);
            }
        }
    }

    int x = nDrawOffX + m_nCurCol - 25;
    const int y = nDrawOffY - nPalHeight - 4 * font->height;

    const int colCurrCol = RGB(255,255,255);
    const int colInavCol = RGB(100,100,100);

    textprintf(bmp, font, x, y - 2, RGB(255,255,0), "Col %d (%d)", m_nColPos + 1, m_nCurCol);

    if (m_nColType == 0)  textprintf(bmp, font, x, y+9,  colCurrCol, "[R] = %.3d", m_pCols[m_nCurCol].r);
    else  textprintf(bmp, font, x, y+9, colInavCol, "[R] = %.3d", m_pCols[m_nCurCol].r);
    if (m_nColType == 1)  textprintf(bmp, font, x, y+18, colCurrCol, "[G] = %.3d", m_pCols[m_nCurCol].g);
    else  textprintf(bmp, font, x, y+18, colInavCol, "[G] = %.3d", m_pCols[m_nCurCol].g);
    if (m_nColType == 2)  textprintf(bmp, font, x, y+27, colCurrCol, "[B] = %.3d", m_pCols[m_nCurCol].b);
    else  textprintf(bmp, font, x, y+27, colInavCol, "[B] = %.3d", m_pCols[m_nCurCol].b);
}
/*-------------------------------------------------------------------------------------*/
int CPalette::FindColInd(int nColPos)
{
    int i = 0;
    int j = 0;
    while (i < m_nCols)
    {
        if (m_pCols[i].bUse)
        {
            if (j == nColPos)          
                return i;
            j++;
        }
        i++;
    }

    if (nColPos < 0)
        return 0;
    else
        return m_nCols - 1;


}
/*-------------------------------------------------------------------------------------*/
void CPalette::HandleMenu(BITMAP * bmp)
{

    if (key[KEY_DOWN] && m_nColPos > m_nMinColPos)
    {
        WAIT(KEY_DOWN);
        m_nColPos--;
        m_nCurCol = FindColInd(m_nColPos);
    }
    if (key[KEY_UP] && m_nColPos < m_nMaxColPos - 1)
    {
        WAIT(KEY_UP);
        m_nColPos++;
        m_nCurCol = FindColInd(m_nColPos);
    }

    if (key[KEY_K])
    {
        WAIT(KEY_K);
        m_bShowKnots = !m_bShowKnots;
    }

    if (key[KEY_INSERT])
    {
        WAIT(KEY_INSERT);
        int n;
        if (m_nCurCol < m_nCols / 2 - 1)
        {
            n = (FindColInd(m_nColPos) + FindColInd(m_nColPos + 1)) / 2;
            m_nColPos++;
        }
        else
        {
            n = (FindColInd(m_nColPos - 1) + FindColInd(m_nColPos)) / 2;
        }

        SetRGB(n, 0, 0, 0, true);
        m_nCurCol = n;
        Make();
    }

    if (key[KEY_DEL])
    {
        WAIT(KEY_DEL);
        if (m_nCurCol != 0 && m_nCurCol != m_nCols - 1)
        {       
            m_pCols[m_nCurCol].bUse = false;
            m_nColPos--;
            m_nMaxColPos--;
            m_nCurCol = FindColInd(m_nColPos);
            Make();
        }
    }

    if (key[KEY_PGUP])
    {
        switch(m_nColType)
        {
            case 0:
                    m_pCols[m_nCurCol].r++;
                    if (m_pCols[m_nCurCol].r > 255) 
                        m_pCols[m_nCurCol].r = 0;
                    break;
            case 1:
                    m_pCols[m_nCurCol].g++;
                    if (m_pCols[m_nCurCol].g > 255) 
                        m_pCols[m_nCurCol].g = 0;
                    break;
            case 2:
                    m_pCols[m_nCurCol].b++;
                    if (m_pCols[m_nCurCol].b > 255) 
                        m_pCols[m_nCurCol].b = 0;
        }
        Make();
    }

    if (key[KEY_PGDN])
    {
        switch(m_nColType)
        {
            case 0:
                    m_pCols[m_nCurCol].r--;
                    if (m_pCols[m_nCurCol].r < 0) 
                        m_pCols[m_nCurCol].r = 255;
                    break;
            case 1:
                    m_pCols[m_nCurCol].g--;
                    if (m_pCols[m_nCurCol].g < 0) 
                        m_pCols[m_nCurCol].g = 255;
                    break;
            case 2:
                    m_pCols[m_nCurCol].b--;
                    if (m_pCols[m_nCurCol].b < 0) 
                        m_pCols[m_nCurCol].b = 255;
        }
        Make();
    }

    if (key[KEY_R]) m_nColType = 0;
    if (key[KEY_G]) m_nColType = 1;
    if (key[KEY_B]) m_nColType = 2;

    m_nPrvCol = m_nCurCol;

    if (m_nColPos != 0 && m_nColPos != m_nMaxColPos - 1)
    {       
        if (key[KEY_LEFT] && m_nCurCol > 1)
        {
            m_nCurCol--;
            if (m_pCols[m_nCurCol].bUse && m_nCurCol > 1) //przeskocz kolor znaczacy
            {
                m_nCurCol--;
                m_nColPos--;
            }
            if (!m_pCols[m_nCurCol].bUse)
            {
                m_pCols[m_nCurCol] = m_pCols[m_nPrvCol];
                m_pCols[m_nPrvCol].bUse = false;
            }
            m_pCols[m_nCurCol].bUse = true;
            Make();
        }
        if (key[KEY_RIGHT] && m_nCurCol < m_nCols - 2)
        {
            m_nCurCol++;
            if (m_pCols[m_nCurCol].bUse && m_nCurCol < m_nCols - 2)
            {
                m_nCurCol++;
                m_nColPos++;
            }
            if (!m_pCols[m_nCurCol].bUse)
            {
                m_pCols[m_nCurCol] = m_pCols[m_nPrvCol];
                m_pCols[m_nPrvCol].bUse = false;
            }
            m_pCols[m_nCurCol].bUse = true;
            Make();
        }
    }
}
/*-------------------------------------------------------------------------------------*/
