#include <allegro.h>
#include "texture.h"
#include "global.h"
#include <fstream>
#include <string>
#include <math.h>
using namespace std;


CTexture::CTexture(int nWidth, int nHeight, int nDepth, int nCols, string strName)
{
    m_pTexture = create_bitmap(nWidth, nHeight);

    m_nWidth  = nWidth;
    m_nHeight = nHeight;
    m_nDepth  = nDepth;
    m_nMode   = MODE_NOISE;

    m_bkp_i = 0;
    m_bkp_j = 0;
    m_bFinished = false;
    m_bChange = true;

    if (nCols)
        m_pPalette = new CPalette(nCols);

    m_strName = strName;
    m_bTurbulence = false;
}

CTexture::~CTexture()
{
    destroy_bitmap(m_pTexture);
    delete m_pPalette;
}

void CTexture::Show(BITMAP * bmp, int nDrawOffsX, int nDrawOffsY, int col)
{
    blit(m_pTexture, bmp, 0, 0, nDrawOffsX, nDrawOffsY, m_nWidth, m_nHeight);
    rect(scrbuf, nDrawOffsX - 1, nDrawOffsY - 1, nDrawOffsX + m_nWidth, nDrawOffsY + m_nHeight, col);
}

int CTexture::Wood(int x, int y)
{
    return (int) (256.0 / sqrt(x *x + y*y));
}
int CTexture::Brick(int x, int y)
{
    int nBrickWidth = 80;
    int nBrickHeight = 30;
    int nMortarWidth = 5;
    int nMortarHeight = 5;
    int nBrickShift = nBrickWidth / 2;

    static int k = 0;
    static int w = 0;
    static int p = 0;

    int n = 0;

    //n = (int) (cos(x/m_fScaleX + PerlinNoise(x/m_fScaleX, y/m_fScaleY, m_fDepth, m_fFreqStart, m_fFreqMod, m_fAmplStart, m_fAmplMod, m_nOctaves) * 40.0) * 255.0, m_bTurbulence);
    if (k == 0)
        nBrickShift = 0;
    else
        nBrickShift = nBrickWidth / 2;

    int xp = (x+nBrickShift) % (nBrickWidth + nMortarWidth);
    int yp = y %  (nBrickHeight + nMortarHeight);
    if (xp < nMortarWidth)
        n = 200 + p * 10;
    if (xp < nMortarWidth)
        p++;
    if (xp == 0)
        p = 0;

    if (yp < nMortarHeight)
        n = 200 + w * 10;

    if (yp < nMortarHeight && x == 255)
        w++;

    if (yp == 0 && x == 255)    
    {
        w = 0;
        k = 1 - k;
    }

    return n;

}
#define RAD(x) (x*3.1415/180.0)
int CTexture::Sinn(int x, int y)
{
    float fx = 256/(float)x;
    float fy = 256/(float)y;

    return (int) (sin(RAD(x)) + cos(RAD(y))* 256.0);
}
void CTexture::Generate()
{
    if (m_nMode == MODE_BMP)
        return;

    if (m_bFinished && !m_bChange)
        return;

    if (m_bChange)
    {
        m_bkp_i = 0;
        m_bkp_j = 0;
    }

    float p;

    bool bDraw = true;

    int i, j;

    for (j = m_bkp_j; j < m_nHeight; j++)
    {
        for (i = m_bkp_i; i < m_nWidth; i++)
        {
            float fx = i / m_fScaleX;
            float fy = j / m_fScaleY;
            float fz = m_fDepth / m_fScaleZ;

            switch (m_nMode)
            {
                case MODE_NOISE:
                    p = PerlinNoise(fx, fy, m_fDepth, m_fFreqStart, m_fFreqMod, m_fAmplStart, m_fAmplMod, m_nOctaves, m_bTurbulence);
                    p = (p + 1.0) * 0.5;
                    //if (p>1.0) p=p/3.0;
                    //p = Clamp(p, 0.0, 1.0);
                    break;
                case MODE_WOOD:
                    p = PerlinNoise(fx, fy, m_fDepth, m_fFreqStart, m_fFreqMod, m_fAmplStart, m_fAmplMod, m_nOctaves, m_bTurbulence);
                    /*p = (p + 1.0) * 0.5;
                    p *= 10;
                    p = p - (int) p; */
                    p = cos(p*180.0/3.1415);
                    //p = Clamp(p, 0.0, 1.0);
                    break;
                case MODE_MARBLE:
                    p = PerlinNoise(fx, fy, m_fDepth, m_fFreqStart, m_fFreqMod, m_fAmplStart, m_fAmplMod, m_nOctaves, m_bTurbulence);
                    //  p = cos(fx + PerlinNoise(fx, fy, m_fDepth, m_fFreqStart, m_fFreqMod, m_fAmplStart, m_fAmplMod, m_nOctaves) * 40.0);
                    if (p<0) p=-p;
                    p = sqrt(p);
                    break;

                case MODE_WOOD2:
                    //t = PerlinNoise(fx, fy, m_fDepth, m_fFreqStart, m_fFreqMod, m_fAmplStart, m_fAmplMod, m_nOctaves);

/*                    x = i / (float)m_nWidth - 0.5;
                    y = j / (float)m_nHeight - 0.5;
                    p = sqrt(x*x + y*y);*/
                    //n = wood_f(i, j);
                    
                    p = 255.0 / (float) Brick(i,j);

                    //p = 255.0 / (float) Sinn(i,j);

                    //_putpixel32(m_pTexture, i, j, RGB(155,255,255));                        
                    //_putpixel32(m_pTexture, i, j, m_pPalette->GetRGB(n));
                    break;

                case MODE_PATTERN:
                       
                    p = i / (float) m_nWidth;
                    break;
                default:
                    bDraw = false;
            }

            if (bDraw)
            {           
                int n = (int) (p * (float) (m_pPalette->m_nCols - 1)) & (m_pPalette->m_nCols - 1);
                _putpixel32(m_pTexture, i, j, m_pPalette->GetRGB(n));
            }
        }
        if (keypressed()) break;
    }

    if (j != m_nHeight)
    {
        m_bFinished = false;
        m_bkp_i = 0;
        m_bkp_j = j;
    }
    else
    {
        m_bFinished = true;
        m_bChange = false;
    }
}

void CTexture::Set(CTexture * tex)
{
    blit(tex->m_pTexture, m_pTexture, 0, 0, 0, 0, m_nWidth, m_nHeight);
}

void CTexture::Add(CTexture * tex)
{
    for (int i = 0; i < m_nWidth; i++)
        for (int j = 0; j < m_nHeight; j++)
        {
            int p0 = _getpixel32(m_pTexture, i, j);
            int p1 = _getpixel32(tex->m_pTexture, i, j);

            int r0 = getr(p0); int r1 = getr(p1);
            int g0 = getg(p0); int g1 = getg(p1);
            int b0 = getb(p0); int b1 = getg(p1);

            int r = r0 + r1;
            int g = g0 + g1;
            int b = b0 + b1;

            if (r > 255) r = 255;
            if (g > 255) g = 255;
            if (b > 255) b = 255;


            _putpixel32(m_pTexture, i, j, makecol(r, g, b));
        }
}

void CTexture::Mul(CTexture * tex)
{
    for (int i = 0; i < m_nWidth; i++)
        for (int j = 0; j < m_nHeight; j++)
        {
            int p0 = _getpixel32(m_pTexture, i, j);
            int p1 = _getpixel32(tex->m_pTexture, i, j);

            int r0 = getr(p0); int r1 = getr(p1);
            int g0 = getg(p0); int g1 = getg(p1);
            int b0 = getb(p0); int b1 = getg(p1);

            int r = sqrt(r0 * r1);
            int g = sqrt(g0 * g1);
            int b = sqrt(b0 * b1);

            if (r > 255) r = 255;
            if (g > 255) g = 255;
            if (b > 255) b = 255;

            _putpixel32(m_pTexture, i, j, makecol(r, g, b));
        }
}

inline float f_characteristic_marble(float x)
{
    return x;
}
inline float f_characteristic_wood(float x, float y)
{
    x = x - 0.5;
    y = y - 0.5;
    return sqrt(x * x + y * y);
}

void CTexture::Mix(CTexture * tex)
{
/*    if (m_bFinished && !m_bChange)
        return;

    if (m_bChange)
    {
        m_bkp_i = 0;
        m_bkp_j = 0;
    }*/
    int i, j;

    for (i = m_bkp_i; i < m_nWidth; i++)
    {
        for (j = m_bkp_j; j < m_nHeight; j++)
        {
            float fx = i / m_fScaleX;
            float fy = j / m_fScaleY;

            float p = PerlinNoise(fx, fy, m_fDepth, m_fFreqStart, m_fFreqMod, m_fAmplStart, m_fAmplMod, m_nOctaves, m_bTurbulence);
            //p = Clamp((p + 1.0) * 0.5, 0.0, 1.0);

            int x = ((float)i/255.0 + p);
            int y = ((float)j/255.0 + p);
//            int p0 =  (int) (f_characteristic_wood(x, y) * 255.0);
            int p0 =  (int) (f_characteristic_marble(i / 255.0 + p) * 255.0);
                //_getpixel32(tex->m_pTexture, (int) (p * 255.0), j);//(int) (j * p * 1.0));

            _putpixel32(m_pTexture, i, j, tex->m_pPalette->GetRGB(p0));
        }
        if (keypressed()) break;
    }
/*    if (j != m_nHeight)
    {
        m_bFinished = false;
        m_bkp_i = 0;
        m_bkp_j = j;
    }
    else
    {
        m_bFinished = true;
        m_bChange = false;
    }*/
}


int CTexture::Save(string strFileName)
{
    if (strFileName == "")
        strFileName = m_strName;

    ofstream file(strFileName.c_str(), ios::out);

    if (!file)
        return -1;

    file << "Mode=" << m_nMode << endl;
    file << "Octaves=" << m_nOctaves << endl;
    file << "AmplStart=" <<  m_fAmplStart << endl;
    file << "AmplMod=" << m_fAmplMod << endl;
    file << "FreqStart=" << m_fFreqStart << endl;
    file << "FreqMod=" << m_fFreqMod << endl;
    file << "Depth=" << m_fDepth << endl;
    file << "ScaleX=" << m_fScaleX << endl;
    file << "ScaleY=" << m_fScaleY << endl;
    file << "NumOfCols=" << m_pPalette->m_nCols << endl;
    file << "Colors" << endl;
    for (int i = 0; i < m_pPalette->m_nCols; i++)
    {
        if (m_pPalette->m_pCols[i].bUse)
            file << (int) i << ","
                 << (int) m_pPalette->m_pCols[i].r << ","
                 << (int) m_pPalette->m_pCols[i].g << ","
                 << (int) m_pPalette->m_pCols[i].b << endl;
    }


    file.close();
    return 0;
}

int CTexture::Load(string strFileName)
{
    if (strFileName == "")
        strFileName = m_strName;

    ifstream file(strFileName.c_str(), ios::in);

    if (!file)
        return -1;

    string strLine;

    while (!file.eof())
    {
        getline(file, strLine, '\n');

    }
    

    file.close();

    return 0;
}